@@ -1,5 +1,4 @@
 # -*- coding: utf-8 -*-
-
 # Copyright 2020 Google LLC
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -14,16 +13,15 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-
 import warnings
-from typing import Awaitable, Callable, Dict, Optional, Sequence, Tuple
+from typing import Awaitable, Callable, Dict, Optional, Sequence, Tuple, Union
 
 from google.api_core import gapic_v1  # type: ignore
 from google.api_core import grpc_helpers_async  # type: ignore
 from google.api_core import operations_v1  # type: ignore
-from google import auth  # type: ignore
-from google.auth import credentials  # type: ignore
+from google.auth import credentials as ga_credentials  # type: ignore
 from google.auth.transport.grpc import SslCredentials  # type: ignore
+import packaging.version
 
 import grpc  # type: ignore
 from grpc.experimental import aio  # type: ignore
@@ -35,11 +33,10 @@ from google.cloud.artifactregistry_v1beta2.types import repository as gda_reposi
 from google.cloud.artifactregistry_v1beta2.types import tag
 from google.cloud.artifactregistry_v1beta2.types import tag as gda_tag
 from google.cloud.artifactregistry_v1beta2.types import version
-from google.iam.v1 import iam_policy_pb2 as iam_policy  # type: ignore
-from google.iam.v1 import policy_pb2 as policy  # type: ignore
-from google.longrunning import operations_pb2 as operations  # type: ignore
-from google.protobuf import empty_pb2 as empty  # type: ignore
-
+from google.iam.v1 import iam_policy_pb2  # type: ignore
+from google.iam.v1 import policy_pb2  # type: ignore
+from google.longrunning import operations_pb2  # type: ignore
+from google.protobuf import empty_pb2  # type: ignore
 from .base import ArtifactRegistryTransport, DEFAULT_CLIENT_INFO
 from .grpc import ArtifactRegistryGrpcTransport
 
@@ -76,7 +73,7 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
     def create_channel(
         cls,
         host: str = "artifactregistry.googleapis.com",
-        credentials: credentials.Credentials = None,
+        credentials: ga_credentials.Credentials = None,
         credentials_file: Optional[str] = None,
         scopes: Optional[Sequence[str]] = None,
         quota_project_id: Optional[str] = None,
@@ -84,7 +81,7 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
     ) -> aio.Channel:
         """Create and return a gRPC AsyncIO channel object.
         Args:
-            address (Optional[str]): The host for the channel to use.
+            host (Optional[str]): The host for the channel to use.
             credentials (Optional[~.Credentials]): The
                 authorization credentials to attach to requests. These
                 credentials identify this application to the service. If
@@ -103,13 +100,15 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
         Returns:
             aio.Channel: A gRPC AsyncIO channel object.
         """
-        scopes = scopes or cls.AUTH_SCOPES
+
+        self_signed_jwt_kwargs = cls._get_self_signed_jwt_kwargs(host, scopes)
+
         return grpc_helpers_async.create_channel(
             host,
             credentials=credentials,
             credentials_file=credentials_file,
-            scopes=scopes,
             quota_project_id=quota_project_id,
+            **self_signed_jwt_kwargs,
             **kwargs,
         )
 
@@ -117,7 +116,7 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
         self,
         *,
         host: str = "artifactregistry.googleapis.com",
-        credentials: credentials.Credentials = None,
+        credentials: ga_credentials.Credentials = None,
         credentials_file: Optional[str] = None,
         scopes: Optional[Sequence[str]] = None,
         channel: aio.Channel = None,
@@ -131,7 +130,8 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
         """Instantiate the transport.
 
         Args:
-            host (Optional[str]): The hostname to connect to.
+            host (Optional[str]):
+                 The hostname to connect to.
             credentials (Optional[google.auth.credentials.Credentials]): The
                 authorization credentials to attach to requests. These
                 credentials identify the application to the service; if none
@@ -162,10 +162,10 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
                 ignored if ``channel`` or ``ssl_channel_credentials`` is provided.
             quota_project_id (Optional[str]): An optional project to use for billing
                 and quota.
-            client_info (google.api_core.gapic_v1.client_info.ClientInfo):	
-                The client info used to send a user-agent string along with	
-                API requests. If ``None``, then default info will be used.	
-                Generally, you only need to set this if you're developing	
+            client_info (google.api_core.gapic_v1.client_info.ClientInfo):
+                The client info used to send a user-agent string along with
+                API requests. If ``None``, then default info will be used.
+                Generally, you only need to set this if you're developing
                 your own client library.
 
         Raises:
@@ -174,7 +174,10 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
           google.api_core.exceptions.DuplicateCredentialArgs: If both ``credentials``
               and ``credentials_file`` are passed.
         """
+        self._grpc_channel = None
         self._ssl_channel_credentials = ssl_channel_credentials
+        self._stubs: Dict[str, Callable] = {}
+        self._operations_client = None
 
         if api_mtls_endpoint:
             warnings.warn("api_mtls_endpoint is deprecated", DeprecationWarning)
@@ -182,89 +185,58 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
             warnings.warn("client_cert_source is deprecated", DeprecationWarning)
 
         if channel:
-            # Sanity check: Ensure that channel and credentials are not both
-            # provided.
+            # Ignore credentials if a channel was passed.
             credentials = False
-
             # If a channel was explicitly provided, set it.
             self._grpc_channel = channel
             self._ssl_channel_credentials = None
-        elif api_mtls_endpoint:
-            host = (
-                api_mtls_endpoint
-                if ":" in api_mtls_endpoint
-                else api_mtls_endpoint + ":443"
-            )
-
-            if credentials is None:
-                credentials, _ = auth.default(
-                    scopes=self.AUTH_SCOPES, quota_project_id=quota_project_id
-                )
-
-            # Create SSL credentials with client_cert_source or application
-            # default SSL credentials.
-            if client_cert_source:
-                cert, key = client_cert_source()
-                ssl_credentials = grpc.ssl_channel_credentials(
-                    certificate_chain=cert, private_key=key
-                )
-            else:
-                ssl_credentials = SslCredentials().ssl_credentials
-
-            # create a new channel. The provided one is ignored.
-            self._grpc_channel = type(self).create_channel(
-                host,
-                credentials=credentials,
-                credentials_file=credentials_file,
-                ssl_credentials=ssl_credentials,
-                scopes=scopes or self.AUTH_SCOPES,
-                quota_project_id=quota_project_id,
-                options=[
-                    ("grpc.max_send_message_length", -1),
-                    ("grpc.max_receive_message_length", -1),
-                ],
-            )
-            self._ssl_channel_credentials = ssl_credentials
         else:
-            host = host if ":" in host else host + ":443"
+            if api_mtls_endpoint:
+                host = api_mtls_endpoint
 
-            if credentials is None:
-                credentials, _ = auth.default(
-                    scopes=self.AUTH_SCOPES, quota_project_id=quota_project_id
-                )
+                # Create SSL credentials with client_cert_source or application
+                # default SSL credentials.
+                if client_cert_source:
+                    cert, key = client_cert_source()
+                    self._ssl_channel_credentials = grpc.ssl_channel_credentials(
+                        certificate_chain=cert, private_key=key
+                    )
+                else:
+                    self._ssl_channel_credentials = SslCredentials().ssl_credentials
 
-            if client_cert_source_for_mtls and not ssl_channel_credentials:
-                cert, key = client_cert_source_for_mtls()
-                self._ssl_channel_credentials = grpc.ssl_channel_credentials(
-                    certificate_chain=cert, private_key=key
-                )
+            else:
+                if client_cert_source_for_mtls and not ssl_channel_credentials:
+                    cert, key = client_cert_source_for_mtls()
+                    self._ssl_channel_credentials = grpc.ssl_channel_credentials(
+                        certificate_chain=cert, private_key=key
+                    )
 
-            # create a new channel. The provided one is ignored.
-            self._grpc_channel = type(self).create_channel(
-                host,
-                credentials=credentials,
-                credentials_file=credentials_file,
-                ssl_credentials=self._ssl_channel_credentials,
-                scopes=scopes or self.AUTH_SCOPES,
-                quota_project_id=quota_project_id,
-                options=[
-                    ("grpc.max_send_message_length", -1),
-                    ("grpc.max_receive_message_length", -1),
-                ],
-            )
-
-        # Run the base constructor.
+        # The base transport sets the host, credentials and scopes
         super().__init__(
             host=host,
             credentials=credentials,
             credentials_file=credentials_file,
-            scopes=scopes or self.AUTH_SCOPES,
+            scopes=scopes,
             quota_project_id=quota_project_id,
             client_info=client_info,
         )
 
-        self._stubs = {}
-        self._operations_client = None
+        if not self._grpc_channel:
+            self._grpc_channel = type(self).create_channel(
+                self._host,
+                credentials=self._credentials,
+                credentials_file=credentials_file,
+                scopes=self._scopes,
+                ssl_credentials=self._ssl_channel_credentials,
+                quota_project_id=quota_project_id,
+                options=[
+                    ("grpc.max_send_message_length", -1),
+                    ("grpc.max_receive_message_length", -1),
+                ],
+            )
+
+        # Wrap messages. This must be done after self._grpc_channel exists
+        self._prep_wrapped_messages(client_info)
 
     @property
     def grpc_channel(self) -> aio.Channel:
@@ -351,7 +323,7 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
     def create_repository(
         self,
     ) -> Callable[
-        [gda_repository.CreateRepositoryRequest], Awaitable[operations.Operation]
+        [gda_repository.CreateRepositoryRequest], Awaitable[operations_pb2.Operation]
     ]:
         r"""Return a callable for the create repository method over gRPC.
 
@@ -373,7 +345,7 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
             self._stubs["create_repository"] = self.grpc_channel.unary_unary(
                 "/google.devtools.artifactregistry.v1beta2.ArtifactRegistry/CreateRepository",
                 request_serializer=gda_repository.CreateRepositoryRequest.serialize,
-                response_deserializer=operations.Operation.FromString,
+                response_deserializer=operations_pb2.Operation.FromString,
             )
         return self._stubs["create_repository"]
 
@@ -409,7 +381,7 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
     def delete_repository(
         self,
     ) -> Callable[
-        [repository.DeleteRepositoryRequest], Awaitable[operations.Operation]
+        [repository.DeleteRepositoryRequest], Awaitable[operations_pb2.Operation]
     ]:
         r"""Return a callable for the delete repository method over gRPC.
 
@@ -432,7 +404,7 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
             self._stubs["delete_repository"] = self.grpc_channel.unary_unary(
                 "/google.devtools.artifactregistry.v1beta2.ArtifactRegistry/DeleteRepository",
                 request_serializer=repository.DeleteRepositoryRequest.serialize,
-                response_deserializer=operations.Operation.FromString,
+                response_deserializer=operations_pb2.Operation.FromString,
             )
         return self._stubs["delete_repository"]
 
@@ -493,7 +465,7 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
     @property
     def delete_package(
         self,
-    ) -> Callable[[package.DeletePackageRequest], Awaitable[operations.Operation]]:
+    ) -> Callable[[package.DeletePackageRequest], Awaitable[operations_pb2.Operation]]:
         r"""Return a callable for the delete package method over gRPC.
 
         Deletes a package and all of its versions and tags.
@@ -514,7 +486,7 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
             self._stubs["delete_package"] = self.grpc_channel.unary_unary(
                 "/google.devtools.artifactregistry.v1beta2.ArtifactRegistry/DeletePackage",
                 request_serializer=package.DeletePackageRequest.serialize,
-                response_deserializer=operations.Operation.FromString,
+                response_deserializer=operations_pb2.Operation.FromString,
             )
         return self._stubs["delete_package"]
 
@@ -575,7 +547,7 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
     @property
     def delete_version(
         self,
-    ) -> Callable[[version.DeleteVersionRequest], Awaitable[operations.Operation]]:
+    ) -> Callable[[version.DeleteVersionRequest], Awaitable[operations_pb2.Operation]]:
         r"""Return a callable for the delete version method over gRPC.
 
         Deletes a version and all of its content. The
@@ -596,7 +568,7 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
             self._stubs["delete_version"] = self.grpc_channel.unary_unary(
                 "/google.devtools.artifactregistry.v1beta2.ArtifactRegistry/DeleteVersion",
                 request_serializer=version.DeleteVersionRequest.serialize,
-                response_deserializer=operations.Operation.FromString,
+                response_deserializer=operations_pb2.Operation.FromString,
             )
         return self._stubs["delete_version"]
 
@@ -753,7 +725,9 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
         return self._stubs["update_tag"]
 
     @property
-    def delete_tag(self) -> Callable[[tag.DeleteTagRequest], Awaitable[empty.Empty]]:
+    def delete_tag(
+        self,
+    ) -> Callable[[tag.DeleteTagRequest], Awaitable[empty_pb2.Empty]]:
         r"""Return a callable for the delete tag method over gRPC.
 
         Deletes a tag.
@@ -772,14 +746,14 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
             self._stubs["delete_tag"] = self.grpc_channel.unary_unary(
                 "/google.devtools.artifactregistry.v1beta2.ArtifactRegistry/DeleteTag",
                 request_serializer=tag.DeleteTagRequest.serialize,
-                response_deserializer=empty.Empty.FromString,
+                response_deserializer=empty_pb2.Empty.FromString,
             )
         return self._stubs["delete_tag"]
 
     @property
     def set_iam_policy(
         self,
-    ) -> Callable[[iam_policy.SetIamPolicyRequest], Awaitable[policy.Policy]]:
+    ) -> Callable[[iam_policy_pb2.SetIamPolicyRequest], Awaitable[policy_pb2.Policy]]:
         r"""Return a callable for the set iam policy method over gRPC.
 
         Updates the IAM policy for a given resource.
@@ -797,15 +771,15 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
         if "set_iam_policy" not in self._stubs:
             self._stubs["set_iam_policy"] = self.grpc_channel.unary_unary(
                 "/google.devtools.artifactregistry.v1beta2.ArtifactRegistry/SetIamPolicy",
-                request_serializer=iam_policy.SetIamPolicyRequest.SerializeToString,
-                response_deserializer=policy.Policy.FromString,
+                request_serializer=iam_policy_pb2.SetIamPolicyRequest.SerializeToString,
+                response_deserializer=policy_pb2.Policy.FromString,
             )
         return self._stubs["set_iam_policy"]
 
     @property
     def get_iam_policy(
         self,
-    ) -> Callable[[iam_policy.GetIamPolicyRequest], Awaitable[policy.Policy]]:
+    ) -> Callable[[iam_policy_pb2.GetIamPolicyRequest], Awaitable[policy_pb2.Policy]]:
         r"""Return a callable for the get iam policy method over gRPC.
 
         Gets the IAM policy for a given resource.
@@ -823,8 +797,8 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
         if "get_iam_policy" not in self._stubs:
             self._stubs["get_iam_policy"] = self.grpc_channel.unary_unary(
                 "/google.devtools.artifactregistry.v1beta2.ArtifactRegistry/GetIamPolicy",
-                request_serializer=iam_policy.GetIamPolicyRequest.SerializeToString,
-                response_deserializer=policy.Policy.FromString,
+                request_serializer=iam_policy_pb2.GetIamPolicyRequest.SerializeToString,
+                response_deserializer=policy_pb2.Policy.FromString,
             )
         return self._stubs["get_iam_policy"]
 
@@ -832,8 +806,8 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
     def test_iam_permissions(
         self,
     ) -> Callable[
-        [iam_policy.TestIamPermissionsRequest],
-        Awaitable[iam_policy.TestIamPermissionsResponse],
+        [iam_policy_pb2.TestIamPermissionsRequest],
+        Awaitable[iam_policy_pb2.TestIamPermissionsResponse],
     ]:
         r"""Return a callable for the test iam permissions method over gRPC.
 
@@ -853,8 +827,8 @@ class ArtifactRegistryGrpcAsyncIOTransport(ArtifactRegistryTransport):
         if "test_iam_permissions" not in self._stubs:
             self._stubs["test_iam_permissions"] = self.grpc_channel.unary_unary(
                 "/google.devtools.artifactregistry.v1beta2.ArtifactRegistry/TestIamPermissions",
-                request_serializer=iam_policy.TestIamPermissionsRequest.SerializeToString,
-                response_deserializer=iam_policy.TestIamPermissionsResponse.FromString,
+                request_serializer=iam_policy_pb2.TestIamPermissionsRequest.SerializeToString,
+                response_deserializer=iam_policy_pb2.TestIamPermissionsResponse.FromString,
             )
         return self._stubs["test_iam_permissions"]
 
